"""
Asynchronous client.

Abstract base class for other asynchronous clients such as WebSockets, ZeroMQ
and aiohttp.
"""
from abc import ABCMeta, abstractmethod

from .client import Client
from .request import Request, Notification
from .prepared_request import PreparedRequest


class AsyncClient(Client, metaclass=ABCMeta):
    @abstractmethod
    async def send_message(self, request, **kwargs):
        """(Abstract)"""

    async def send(self, request, **kwargs):
        request = PreparedRequest(request)
        self.prepare_request(request, **kwargs)
        self.log_request(request, request.log_extra, request.log_format)
        return await self.send_message(request, **kwargs)

    async def notify(self, method_name, *args, **kwargs):
        return await self.send(Notification(method_name, *args, **kwargs))

    async def request(self, method_name, *args, **kwargs):
        return await self.send(Request(method_name, *args, **kwargs))



"""Prepared request."""
import json


class PreparedRequest(str):
    """
    Encapsulate and validate a request prior to sending.

    This object can be passed around subsequent methods. Subclasses of Client
    can add to it, or set the log format/extra info to include.

    Converts some non-string requests to string.
    """

    def __new__(cls, request):
        # Convert a list of strings, to one string
        if isinstance(request, list) and all(
            isinstance(i, str) for i in request
        ):
            request = "[{}]".format(", ".join(request))
        # Convert a json-serializable object (dict or list) to a string
        if not isinstance(request, str):
            request = json.dumps(request)
        # Should end up with a string
        assert isinstance(request, str)
        return str.__new__(cls, request)

    def __init__(self, request):
        super().__init__()
        #: Extra details used in log entry, can be set by clients in
        #: prepare_request
        self.log_extra = None
        self.log_format = None



"""ZeroMQ Client"""
import zmq
import zmq.asyncio

from ..async_client import AsyncClient


class ZeroMQAsyncClient(AsyncClient):
    """
    :param endpoint:
    :param socket_type:
    :param *args: Passed through to Client class.
    :param **kwargs: Passed through to Client class.
    """
    def __init__(self, endpoint, *args, socket_type=zmq.REQ, **kwargs):
        super().__init__(endpoint, *args, **kwargs)
        self.context = zmq.asyncio.Context()
        self.socket = self.context.socket(socket_type)
        self.socket.connect(endpoint)

    async def send_message(self, request, **kwargs):
        await self.socket.send_multipart((request.encode(),))
        response = await self.socket.recv_multipart()
        return self.process_response(response[0].decode())



"""
Exceptions
**********

These exceptions are raised when processing responses from the server. For
example, if the response was garbage and could not be parsed,
:class:`ParseResponseError <jsonrpcclient.exceptions.ParseResponseError>` is
raised.

To handle them, use a try-block when calling ``notify`` or ``request``::

    try:
        client.notify('go')
    except JsonRpcClientError as e:
        print(str(e))
"""


class JsonRpcClientError(Exception):
    """Base class for the other exceptions."""

    pass


class ParseResponseError(JsonRpcClientError):
    """The response was not valid JSON."""

    def __init__(self):
        super().__init__("The response was not valid JSON")


class ReceivedErrorResponseError(JsonRpcClientError):
    """
    Raised if a single JSON-RPC `error response object
    <http://www.jsonrpc.org/specification#error_object>`_ is received.

    *This is not raised for batch requests.*

    This error means one of two things:

    1. There was a problem with the request.
    2. There was a problem with on server end.

    To see information about the error, catch the exception:

    .. code-block:: python
        :emphasize-lines: 5-6

        from jsonrpcclient.exceptions import (
            JsonRpcClientError, ReceivedErrorResponseError)
        try:
            client.notify('go')
        except ReceivedErrorResponseError as e:
            print(e.code, e.message, e.data)
        except JsonRpcClientError as e:
            print(str(e))
    """

    def __init__(self, code, message, data):
        """
        :param code: The JSON-RPC ``code`` value received.
        :param message: The JSON-RPC ``message`` value received.
        :param data: The JSON-RPC ``data`` value received.
        """
        super().__init__(message)
        #: The JSON-RPC status code. (See `status codes
        #: <http://www.jsonrpc.org/specification#error_object>`_.)
        self.code = code
        #: A one-line message explaining the error.
        self.message = message
        #: Extra information about the error, if given.
        self.data = data


class ReceivedNon2xxResponseError(JsonRpcClientError):
    """The response was not valid JSON."""

    def __init__(self, status_code):
        super().__init__("Received {} status code".format(status_code))
        self.code = status_code



"""
Tornado Client
**************

Represents an endpoint to communicate with using Tornado asynchronous HTTP
client::

    from tornado import ioloop

    async def test():
        client = TornadoClient('http://example.com/api')
        result = await yield client.some_method(42)
        print(result)

    ioloop.IOLoop.instance().run_sync(test)
"""
from functools import partial

from tornado.httpclient import AsyncHTTPClient

from ..async_client import AsyncClient
from ..exceptions import ReceivedNon2xxResponseError


class TornadoClient(AsyncClient):
    """
    :param endpoint: The server address.
    :param async_http_client_class: Tornado asynchronous HTTP client class.
    :param kwargs: Keyword arguments to pass to the client initialiser.
    """

    DEFAULT_HEADERS = {"Content-Type": "application/json", "Accept": "application/json"}

    def __init__(self, *args, async_http_client=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.http_client = async_http_client or AsyncHTTPClient()

    async def send_message(self, request, **kwargs):
        """
        Transport the message to the server and return the response.

        :param request: The JSON-RPC request string.
        :param kwargs: Keyword arguments to the Tornado request.
        :return: The response (a string for requests, None for notifications).
        """
        headers = dict(self.DEFAULT_HEADERS)
        headers.update(kwargs.pop("headers", {}))

        response = await self.http_client.fetch(
            self.endpoint, method="POST", body=request, headers=headers, **kwargs
        )

        if not 200 <= response.code <= 299:
            raise ReceivedNon2xxResponseError(response.status_code)

        # Note: Tornado adds it's own logger handlers, so the following log format isn't
        # used, unless Tornado's handlers are disabled.
        return self.process_response(
            response.body.decode(),
            log_extra={"http_code": response.code, "http_reason": response.reason},
            log_format="<-- %(message)s (%(http_code)s %(http_reason)s)",
        )



"""
An HTTP client.

For example::

    HTTPClient('http://example.com/api').request('go')

Uses the `Requests <http://docs.python-requests.org/en/master/>`_ library.
"""
from requests import Request, Session

from ..client import Client
from ..exceptions import ReceivedNon2xxResponseError


class HTTPClient(Client):
    """Defines an HTTP client"""

    # The default HTTP header
    DEFAULT_HEADERS = {"Content-Type": "application/json", "Accept": "application/json"}

    def __init__(self, *args, **kwargs):
        """
        :param endpoint: The server address.
        :param **kwargs: Pased through to the Client class.
        """
        super().__init__(*args, **kwargs)
        # Make use of Requests' sessions feature
        self.session = Session()
        self.session.headers.update(self.DEFAULT_HEADERS)

    def prepare_request(self, request, **kwargs):
        """
        Prepare the request for sending.

        :param request: The JSON-RPC request (a PreparedRequest object)
        :param kwargs: Configuration for just this request
        :return: None
        """
        # Use the Requests library to prepare the request based on the session
        # configuration
        req = Request(method="POST", url=self.endpoint, data=request, **kwargs)
        request.prepped = self.session.prepare_request(req)
        # Include the http headers in log extra. Will not have effect unless
        # user configures the log format
        request.log_extra = {"http_headers": request.prepped.headers}

    def send_message(self, request, **kwargs):
        """
        Transport the message to the server and return the response.

        :param request: The JSON-RPC request string.
        :param kwargs: Passed on to the requests lib's send function, for last
            minute configuration (Optional)
        :return: The JSON-RPC response.
        :rtype: A string for requests, None for notifications.
        :raise requests.exceptions.RequestException:
            Raised by the requests module in the event of a communications
            error.
        """
        # Send the message with Requests, passing any final config options
        response = self.session.send(request.prepped, **kwargs)
        if not 200 <= response.status_code <= 299:
            raise ReceivedNon2xxResponseError(response.status_code)
        # Give some extra information to include in the response log entry
        return self.process_response(
            response.text,
            log_extra={
                "http_code": response.status_code,
                "http_reason": response.reason,
            },
            log_format="<-- %(message)s (%(http_code)s %(http_reason)s)",
        )


def notify(
    endpoint,
    method,
    *args,
    trim_log_values=False,
    validate_against_schema=True,
    **kwargs
):
    """
    A convenience function. Instantiates and executes a HTTPClient request, then
    throws it away.
    """
    return HTTPClient(
        endpoint,
        trim_log_values=trim_log_values,
        validate_against_schema=validate_against_schema,
    ).notify(method, *args, **kwargs)


def request(
    endpoint,
    method,
    *args,
    id_generator=None,
    trim_log_values=False,
    validate_against_schema=True,
    **kwargs
):
    """
    A convenience function. Instantiates and executes a HTTPClient request, then
    throws it away.
    """
    return HTTPClient(
        endpoint,
        id_generator=id_generator,
        trim_log_values=trim_log_values,
        validate_against_schema=validate_against_schema,
    ).request(method, *args, **kwargs)



"""
Websockets client.

http://websockets.readthedocs.io/
"""
from ..async_client import AsyncClient


class WebSocketsClient(AsyncClient):
    def __init__(self, socket, *args, **kwargs):
        """
        :param endpoint:
        :param socket_type:
        :param *args: Passed through to Client class.
        :param **kwargs: Passed through to Client class.
        """
        super().__init__(*args, **kwargs)
        self.socket = socket

    async def send_message(self, request, **kwargs):
        await self.socket.send(request)
        response = await self.socket.recv()
        return self.process_response(response)



"""Logging"""
import logging
import json


def configure_logger(logger, fmt):
    """
    Set up a logger, if no handler has been configured for it.

    Used by the log function below.
    """
    if logger.level == logging.NOTSET:
        logger.setLevel(logging.INFO)
    if not logging.root.handlers and not logger.handlers:
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(fmt=fmt))
        logger.addHandler(handler)


def _trim_string(message):
    longest_string = 30

    if len(message) > longest_string:
        prefix_len = int(longest_string / 3)
        suffix_len = prefix_len
        return message[:prefix_len] + "..." + message[-suffix_len:]

    return message


def _trim_values(message_obj):
    result = {}
    longest_list = 30
    for k, val in message_obj.items():
        if isinstance(val, str):
            result[k] = _trim_string(val)
        elif isinstance(val, list) and len(val) > longest_list:
            prefix_len = int(longest_list / 3)
            suffix_len = prefix_len
            result[k] = val[:prefix_len] + ["..."] + val[-suffix_len:]
        elif isinstance(val, dict):
            result[k] = _trim_values(val)
        else:
            result[k] = val
    return result


def trim_message(message):
    try:
        message_obj = json.loads(message)
        return json.dumps(_trim_values(message_obj))
    except ValueError:
        return _trim_string(str(message))


def log_(logger, level, message, *args, **kwargs):
    """
    Log a message.
    """
    fmt = kwargs.pop("fmt", "%(message)s")
    trim = kwargs.pop("trim", False)
    if trim:
        message = trim_message(message)
    configure_logger(logger, fmt)
    getattr(logger, level)(message, *args, **kwargs)



"""Abstract base class for various clients."""
import json
import logging
import pkgutil
from abc import ABCMeta, abstractmethod

import jsonschema

from . import exceptions, ids
from .log import log_
from .prepared_request import PreparedRequest
from .request import Notification, Request


class Client(metaclass=ABCMeta):
    """
    Protocol-agnostic base class for clients.

    Subclasses should inherit and override ``send_message``.

    :param endpoint: The server address.
    """

    # Request and response logs
    request_log = logging.getLogger(__name__ + ".request")
    response_log = logging.getLogger(__name__ + ".response")

    #: Validate the response message
    validator = jsonschema.Draft4Validator(
        json.loads(pkgutil.get_data(__name__, "response-schema.json").decode("utf-8"))
    )

    def __init__(
        self,
        endpoint,
        id_generator=None,
        trim_log_values=False,
        validate_against_schema=True,
    ):
        """
        :param endpoint: Holds the server address.
        :param id_generator: 
        :param trim_log_values: Log abbreviated versions of requests and responses
        :param validate_against_schema: Validate responses against the JSON-RPC schema.
        """
        self.endpoint = endpoint
        self.id_generator = id_generator
        self.trim_log_values = trim_log_values
        self.validate_against_schema = validate_against_schema

    def log_(self, message, extra, log, level, fmt, trim):
        """Log a request or response"""
        if extra is None:
            extra = {}
        # Add the endpoint to the log entry
        extra.update({"endpoint": self.endpoint})
        # Clean up the message for logging
        if isinstance(message, str):
            message = message.replace("\n", "").replace("  ", " ").replace("{ ", "{")
        log_(log, level, message, extra=extra, fmt=fmt, trim=trim)

    def log_request(self, request, extra=None, fmt=None, trim=False):
        """
        Log a request.

        Should be called by subclasses in :meth:`send_message`, before sending.

        :param request: The JSON-RPC request string.
        :param extra: A dict of extra fields that may be logged.
        """
        if not fmt:
            fmt = "--> %(message)s"
        self.log_(request, extra, self.request_log, "info", fmt, trim)

    def log_response(self, response, extra=None, fmt=None, trim=False):
        """
        Log a response.

        Should be called by subclasses in :meth:`send_message`, after receiving
        the response.

        :param response: The JSON-RPC response string.
        :param extra: A dict of extra fields that may be logged.
        """
        if not fmt:
            fmt = "<-- %(message)s"
        self.log_(response, extra, self.response_log, "info", fmt, trim)

    def prepare_request(self, request, **kwargs):
        """
        Prepare the request if necessary.

        Subclasses can overload to modify the request, or to add extra info to
        the log entry (set the extra attribute).
        """
        pass

    def process_response(self, response, log_extra=None, log_format=None):
        """
        Process the response and return the 'result' portion if present.

        :param response: The JSON-RPC response string to process.
        :return: The response string, or None
        """
        if response:
            # Log the response before processing it
            self.log_response(
                response, log_extra, log_format, trim=self.trim_log_values
            )
            # If it's a json string, parse to object
            if isinstance(response, str):
                try:
                    response = json.loads(response)
                except ValueError:
                    raise exceptions.ParseResponseError()
            # Validate the response against the Response schema (raises
            # jsonschema.ValidationError if invalid)
            if self.validate_against_schema:
                self.validator.validate(response)
            if isinstance(response, list):
                # Batch request - just return the whole response
                return response
            else:
                # If the response was "error", raise to ensure it's handled
                if "error" in response and response["error"] is not None:
                    raise exceptions.ReceivedErrorResponseError(
                        response["error"].get("code"),
                        response["error"].get("message"),
                        response["error"].get("data"),
                    )
                # All was successful, return just the result part
                return response.get("result")
        # No response was given
        return None

    @abstractmethod
    def send_message(self, request, **kwargs):
        """
        Transport the request to the server.

        Override this method in the protocol-specific subclasses.

        :param request: A JSON-RPC request, in dict format.
        :returns: The processed response - for requests, it will be the result
        part of the response, None for notifications, or an entire jsonrpc
        response for batch requests.
        """

    def send(self, request, **kwargs):
        """
        Send a request, passing the whole JSON-RPC request object.

            >>> client.send('{"jsonrpc": "2.0", "method": "ping", "id": 1}')
            --> {"jsonrpc": "2.0", "method": "ping", "id": 1}
            <-- {"jsonrpc": "2.0", "result": "pong", "id": 1}
            'pong'

        :param request: The JSON-RPC request.
        :type request: JSON-encoded string or JSON serializable object
        :param kwargs: Clients can use these to configure an single request
            (separate to configuration of the whole session). For example,
            HTTPClient passes them on to `requests.Session.send()
            <http://docs.python-requests.org/en/master/api/#requests.Session.send>`_.
        :returns: The payload, i.e. the ``result`` part of the response, (or
            ``None`` in the case of a Notification).
        :rtype: A `JSON-decoded object
            <https://docs.python.org/library/json.html#json-to-py-table>`_.
        :raises ParseResponseError:
            The response was not valid JSON.
        :raises ValidationError:
            The response was valid JSON, but not a valid JSON-RPC response
            object.
        :raises ReceivedErrorResponseError:
            The server responded with a JSON-RPC `error object
            <http://www.jsonrpc.org/specification#error_object>`_.
        """
        # Put the raw request into an object which can be passed around the
        # subsequent method calls. Converts the request to a string not already.
        request = PreparedRequest(request)
        # Prepare request, subclasses can override to prepare the request, and
        # set the extra details to include in the log entry
        self.prepare_request(request)
        # Log the request
        self.log_request(
            request, request.log_extra, request.log_format, trim=self.trim_log_values
        )
        # Call abstract method to transport the message, returning either the
        # processed response, or a future which promises to process eventually
        return self.send_message(request, **kwargs)

    def notify(self, method_name, *args, **kwargs):
        """
        Send a JSON-RPC request, without expecting a response.

        :param method_name: The remote procedure's method name.
        :param args: Positional arguments passed to the remote procedure.
        :param kwargs: Keyword arguments passed to the remote procedure.
        :return: The payload (i.e. the ``result`` part of the response).
        """
        return self.send(Notification(method_name, *args, **kwargs))

    def request(self, method_name, *args, **kwargs):
        """
        Send a request by passing the method and arguments.

        This is the main public method.

            >>> client.request('cat', name='Mittens')
            --> {"jsonrpc": "2.0", "method": "cat", "params": {"name": "Mittens"}, "id": 1}
            <-- {"jsonrpc": "2.0", "result": "meow", "id": 1}
            'meow'

        :param method_name: The remote procedure's method name.
        :param args: Positional arguments passed to the remote procedure.
        :param kwargs: Keyword arguments passed to the remote procedure.
        :return: The payload (i.e. the ``result`` part of the response).
        """
        return self.send(
            Request(method_name, *args, id_generator=self.id_generator, **kwargs)
        )

    def __getattr__(self, name):
        """
        This gives us an alternate way to make a request::

            >>> client.cube(3)
            27

        That's the same as saying ``client.request('cube', 3)``.
        """

        def attr_handler(*args, **kwargs):
            """Call self.request from here"""
            return self.request(name, *args, **kwargs)

        return attr_handler



"""
aiohttp client.

http://aiohttp.readthedocs.io/
"""
import async_timeout

from ..async_client import AsyncClient
from ..exceptions import ReceivedNon2xxResponseError


class AiohttpClient(AsyncClient):
    """TODO: rename AiohttpClient to AiohttpClient"""

    def __init__(self, session, endpoint):
        super().__init__(endpoint)
        self.session = session

    async def send_message(self, request):
        with async_timeout.timeout(10):
            async with self.session.post(self.endpoint, data=request) as response:
                if not 200 <= response.status <= 299:
                    raise ReceivedNon2xxResponseError(response.status)
                return self.process_response(
                    await response.text(),
                    log_extra={
                        "http_code": response.status,
                        "http_reason": response.reason,
                    },
                    log_format="<-- %(message)s (%(http_code)s %(http_reason)s)",
                )
